#!/usr/bin/runhaskell

> module Main where

I originally wrote harsh in C, using libnbio, curses, and expat, and doing
everything else by hand. I ported it to Haskell as a learning exercise, mainly
to learn about STM and monads.

harsh is an rss aggregator with a console UI. There are really only three
screens: the menu (showing all the feeds), the feed screen (showing all the
items for a particular feed), and the item screen (showing just one particular
item of a feed.

Throughout the comments, I use first person plural ("we") usually to refer to
the program itself.

> import Control.Monad
> import Control.Concurrent
> import Control.Concurrent.STM
> import qualified Control.Exception as E
> import qualified Data.ByteString.Char8 as B
> import Data.Char
> import Graphics.Vty as V
> import List
> import Maybe
> import Network.URI
> import System.Directory
> import System.Environment
> import System.IO

> import Config
> import Feed

This contains all the information about the current state of the UI. We're
using VTY, which makes some things really easy. However, we want to redraw the
UI either when a feed's state has changed, or when a key has been pressed
(event occurs). So we create a new thread just to poll for events, and stuff
the events in a TChan.

> data UI = UI {
>               vt :: Vty,
>               evchan :: TChan Event,
>               fidx :: Int,
>               nr :: TVar Bool,
>               fds :: [TVar Feed],
>               guids :: [[String]],
>               iidx :: Int,
>               draw :: UI -> IO (),
>               prompt :: Maybe (String, String, (UI -> IO ())),
>               nf :: String,
>               handleEvt :: Event -> UI -> IO ()
>           }

> blank :: Int -> Int -> Image
> blank sx sy = renderFill V.attr ' ' sx sy

> renderPostFill :: Int -> String -> Image
> renderPostFill sx = renderBS V.attr . B.pack . take sx . (++ (repeat ' '))

> createLineFrom :: Eq a => Int -> Maybe a -> ((a, [String]) -> String) -> (a, [String]) -> Image
> createLineFrom sx Nothing  f a = renderPostFill sx . (' ':) . f $ a
> createLineFrom sx (Just i) f a = renderPostFill sx . sel . f $ a
>   where sel = if (i == (fst a)) then ('>':) else (' ':)

> showFeed :: (Feed, [String]) -> String
> showFeed (f, g) = st ++ nw ++ " " ++ tl ++ sp ++ er
>   where st = case (status f) of
>               Updating  -> "."
>               NoError   -> " "
>               otherwise -> "!"
>         nw = if length g >= length (items f) then " " else "N"
>         tl = fromMaybe (show $ uri f) (feedtitle f)
>         sp = take (70 - (length tl)) (repeat ' ')
>         er = case (status f) of
>               NetError  -> "NET"
>               HdrError  -> "HDR"
>               XmlError  -> "XML"
>               RssError  -> "RSS"
>               otherwise -> ""

> drawFeeds :: Int -> Int -> Int -> [(Feed, [String])] -> Image
> drawFeeds sx sy selected feeds =
>   foldl (<->) (blank sx 1) (map (createLineFrom sx sel showFeed) feeds) <-> blankToEnd
>   where sel = Just $ fst (feeds!!selected)
>         blankToEnd = blank sx (sy - (length feeds) - 2)

> drawPrompt :: Int -> Maybe (String, String, a) -> Image
> drawPrompt sx Nothing          = renderFill V.attr ' ' sx 1
> drawPrompt sx (Just (p, e, _)) = renderBS V.attr $ B.pack $ take sx $ p ++ ": " ++ e ++ (repeat ' ')

> menuDraw :: UI -> IO ()
> menuDraw ui = do
>   feeds <- atomically $ mapM readTVar (fds ui)
>   (sx, sy) <- getSize (vt ui)
>   update (vt ui) (pic { pCursor = if isNothing (prompt ui) then NoCursor else Cursor plen sy,
>                         pImage = drawFeeds sx sy (fidx ui) (zip feeds (guids ui))
>                                  <-> drawPrompt sx (prompt ui) })
>   where plen = (\ (a,b,_) -> length a + length b + 2) (fromJust $ prompt ui)

> addFeed :: UI -> IO ()
> addFeed ui = case feed of
>   Nothing -> return (ui { prompt = Nothing,
>                           handleEvt = menuEvt }) >>= drawNRun
>   Just f  -> do
>               var <- newTVarIO f
>               forkIO $ monitorFeed var (nr ui)
>               fds <- atomically $ mapM readTVar (fds ui)
>               writeConfig (fds ++ [f]) ((guids ui) ++ [[]])
>               return (ui { prompt = Nothing,
>                            fds = append var,
>                            handleEvt = menuEvt,
>                            guids = (guids ui) ++ [[]] }) >>= drawNRun
>   where feed = case entered of
>                  "" -> Nothing
>                  _  -> case reads entered of
>                          (n,_):_ -> newFeed (nf ui) n
>                          []      -> Nothing
>         (_, entered, _) = fromJust $ prompt ui
>         append v = (fds ui) ++ [v]

> getIvl :: UI -> IO ()
> getIvl ui | entered == "" = return (ui { prompt = Nothing,
>                                          handleEvt = menuEvt }) >>= drawNRun
>           | otherwise     = return (ui { nf = entered,
>                                          prompt = Just ("Interval in hours to check feed",
>                                                         "", addFeed) }) >>= drawNRun
>   where (_, entered, _) = fromJust $ prompt ui

> promptEvt :: Event -> UI -> IO ()
> promptEvt evt ui = case evt of
>   EvKey (KASCII 'g') [MCtrl] -> return (ui { prompt = Nothing,
>                                              handleEvt = menuEvt }) >>= drawNRun
>   EvKey (KASCII x)   []      -> return (ui { prompt = append x }) >>= drawNRun
>   EvKey (KEnter)     []      -> enter ui
>   _                          -> runUI ui
>   where (p,e,enter) = fromJust $ prompt ui
>         append x    = Just (p, e ++ [x], enter)

> moveFeed :: (Int -> Int) -> UI -> UI
> moveFeed f ui = ui { fidx = max (min newf last) 0 }
>   where newf = f (fidx ui)
>         last = (length (fds ui)) - 1

> reloadCurFeed :: UI -> IO ()
> reloadCurFeed ui = do
>   feed <- atomically $ readTVar ((fds ui)!!(fidx ui))
>   if (status feed) == Updating
>     then return ()
>     else case thread feed of
>       Nothing -> return ()
>       Just th -> E.throwDynTo th Restart

> menuEvt :: Event -> UI -> IO ()
> menuEvt evt ui = case evt of
>   EvKey (KASCII 'q') []      -> shutdown (vt ui)
>   EvKey (KASCII 'a') []      -> return (ui { prompt = Just ("Add feed url", "", getIvl),
>                                              handleEvt = promptEvt }) >>= drawNRun
>   EvKey (KASCII 'j') []      -> return (moveFeed (succ) ui) >>= drawNRun
>   EvKey (KASCII 'k') []      -> return (moveFeed (pred) ui) >>= drawNRun
>   EvKey (KASCII '=') []      -> return (moveFeed (\ _ -> 0) ui) >>= drawNRun
>   EvKey (KASCII '*') []      -> return (moveFeed (\ _ -> length (fds ui)) ui) >>= drawNRun
>   EvKey (KASCII 'r') [MCtrl] -> reloadCurFeed ui >> runUI ui
>   EvKey (KEnter)     []      -> return (ui { draw = feedDraw, handleEvt = feedEvt, iidx = 0 }) >>=
>                                   drawNRun
>   _                          -> runUI ui

> showItem :: (Item, [String]) -> String
> showItem (i, ids) = rd ++ mn ++ cr
>   where rd = if (guid i) `elem` ids then "   " else " N "
>         mn = if isJust (itemtitle i) then fromJust (itemtitle i) else fromJust (itemdesc i)
>         cr = if isJust (author i)
>               then " (" ++ (fromJust (author i)) ++ ")"
>               else ""

> drawItems :: Int -> Int -> Feed -> [String] -> Int -> Image
> drawItems sx sy feed ids cur = (blank sx 1) <-> fddesc <->
>   (foldl (<->) (blank sx 1) (map (createLineFrom sx sel showItem . (flip (,)) ids) (items feed))) <->
>   blankToEnd
>   where fddesc  = renderPostFill sx . ("  " ++) . fromMaybe "" . feeddesc $ feed
>         sel = Just $ items feed!!cur
>         blankToEnd = blank sx (sy - (length (items feed)) - 3)

> feedDraw :: UI -> IO ()
> feedDraw ui = do
>   feed <- atomically $ readTVar $ (fds ui)!!(fidx ui)
>   (sx, sy) <- getSize (vt ui)
>   update (vt ui) (pic { pCursor = NoCursor,
>                         pImage = drawItems sx sy feed ((guids ui)!!(fidx ui)) (iidx ui) })

> addGuid :: [[String]] -> Int -> String -> [[String]]
> addGuid orig idx new = if new `elem` (orig!!idx)
>   then orig
>   else zipWith (++) orig add
>   where add = map (\ n -> if (n == idx) then [new] else []) [0..]

> markItemRead :: (Int -> Int) -> UI -> IO UI
> markItemRead fn ui = do
>   fds <- atomically $ mapM readTVar (fds ui)
>   let fd = fds!!(fidx ui)
>   let upd = ui { guids = addGuid (guids ui) (fidx ui) (guid ((items fd)!!(iidx ui))),
>                  iidx = max (min new (last fd)) 0 }
>   writeConfig (fds) (guids upd)
>   return upd
>     where new     = fn (iidx ui)
>           last fd = (length (items fd)) - 1

> lastItem :: UI -> IO Int
> lastItem ui = do
>   fd <- atomically $ readTVar ((fds ui)!!(fidx ui))
>   return $ (length (items fd)) - 1

> moveItem :: (Int -> Int) -> UI -> IO UI
> moveItem f ui = do
>   last <- lastItem ui
>   let newf = f (iidx ui)
>   return ui { iidx = max (min newf last) 0 }

> nextUnread :: UI -> IO Int
> nextUnread ui = do
>   fd <- atomically $ readTVar ((fds ui)!!(fidx ui))
>   return $ case inds fd of
>     [] -> iidx ui
>     _  -> case bigInds fd of
>             [] -> head $ inds fd
>             _  -> head $ bigInds fd
>   where inds fd = findIndices (\ i -> not $ (guid i) `elem` ((guids ui)!!(fidx ui))) (items fd)
>         bigInds fd = [ x | x <- inds fd, x > iidx ui ]

> feedEvt :: Event -> UI -> IO ()
> feedEvt evt ui = case evt of
>   EvKey (KASCII 'i') []      -> return (ui { draw = menuDraw, handleEvt = menuEvt }) >>= drawNRun
>   EvKey (KASCII 'q') []      -> return (ui { draw = menuDraw, handleEvt = menuEvt }) >>= drawNRun
>   EvKey (KASCII 'j') []      -> moveItem (succ) ui >>= drawNRun
>   EvKey (KASCII 'k') []      -> moveItem (pred) ui >>= drawNRun
>   EvKey (KASCII '=') []      -> moveItem (\ _ -> 0) ui >>= drawNRun
>   EvKey (KASCII '*') []      -> lastItem ui >>= (\ last -> moveItem (\ _ -> last) ui) >>= drawNRun
>   EvKey (KASCII 'i') [MCtrl] -> nextUnread ui >>= (\ last -> moveItem (\ _ -> last) ui) >>= drawNRun
>   EvKey (KASCII 'r') [MCtrl] -> reloadCurFeed ui >> runUI ui
>   EvKey (KASCII 'N') []      -> markItemRead succ ui >>= drawNRun
>   EvKey (KEnter)     []      -> markItemRead id ui >>=
>                                   (\ u -> return u { draw = itemDraw, handleEvt = itemEvt }) >>=
>                                   drawNRun
>   _                          -> runUI ui

> showContent :: Int -> String -> (Image, Int)
> showContent sx s | length s <= sx = (renderPostFill sx s, 1)
>                  | otherwise      = (renderPostFill sx a <-> showRemaining, 1 + countRemain)
>   where (showRemaining, countRemain) = showContent sx b
>         (a, b) = splitAt sx s

> drawItem :: Int -> Int -> Feed -> Int -> Image
> drawItem sx sy feed idx = (blank sx 1) <-> fddesc <-> (blank sx 1) <->
>   createLineFrom sx Nothing showItem (item, gd) <-> (blank sx 1) <->
>   content <-> blankToEnd
>   where fddesc  = renderPostFill sx . ("  " ++) . fromMaybe "" . feeddesc $ feed
>         item = (items feed)!!idx
>         gd   = [guid item]
>         blankToEnd = blank sx (sy - 5 - height)
>         (content, height) = case itemdesc item of
>                               Just s  -> showContent sx ("      " ++ s)
>                               Nothing -> (renderPostFill sx "      no content", 1)

> itemDraw :: UI -> IO ()
> itemDraw ui = do
>   feed <- atomically $ readTVar $ (fds ui)!!(fidx ui)
>   (sx, sy) <- getSize (vt ui)
>   update (vt ui) (pic { pCursor = NoCursor,
>                         pImage = drawItem sx sy feed (iidx ui) })

> itemEvt :: Event -> UI -> IO ()
> itemEvt evt ui = case evt of
>   EvKey (KASCII 'i') []      -> return (ui { draw = feedDraw, handleEvt = feedEvt }) >>= drawNRun
>   EvKey (KASCII 'q') []      -> return (ui { draw = feedDraw, handleEvt = feedEvt }) >>= drawNRun
>   EvKey (KASCII 'j') []      -> moveItem (succ) ui >>= markItemRead id >>= drawNRun
>   EvKey (KASCII 'k') []      -> moveItem (pred) ui >>= markItemRead id >>= drawNRun
>   _                          -> runUI ui

We keep track of which items have been read, by guid. When a feed is refreshed,
we only want to remember the guids that are still in the feed. We don't bother
keeping track of the old guids, nor the items that are no longer present.

XXX we're calling this each time a feed's state changes, and it runs over all
feeds! What we really want is to only do this as necessary. That way we don't
blow away all the read guids when we start up. I've worked around this for now,
but it's a hack.

> updateGuids :: UI -> IO UI
> updateGuids ui = do
>   fs <- atomically $ mapM readTVar (fds ui)
>   return $ ui { guids = updatedGuids fs }
>   where updatedGuids = map filterGuids . zip (guids ui) . map (map guid . items)
>         filterGuids (orig, cur) = if (length cur == 0) then orig else filter (`elem` cur) orig

> drawNRun :: UI -> IO ()
> drawNRun u = (draw u) u >> runUI u

> waitForFeeds :: TVar Bool -> STM (Either () Event)
> waitForFeeds nr = do
>   need <- readTVar nr
>   if need then return () else retry
>   writeTVar nr False
>   return (Left ())

> waitForEvents :: TChan Event -> STM (Either () Event)
> waitForEvents ichan = do
>   evt <- readTChan ichan
>   return $ Right evt

> runUI :: UI -> IO ()
> runUI ui = do
>   eith <- atomically $ waitForFeeds (nr ui) `orElse` waitForEvents (evchan ui)
>   case eith of
>     Left _  -> updateGuids ui >>= drawNRun
>     Right e -> (handleEvt ui) e ui

> pollForEvents :: Vty -> TChan Event -> IO ()
> pollForEvents vt ichan = forever $ do
>   evt <- getEvent vt
>   atomically $ writeTChan ichan evt

> spawnFeed :: TVar Bool -> Feed -> IO (TVar Feed)
> spawnFeed nerd f = do
>   var <- newTVarIO f
>   forkIO $ monitorFeed var nerd
>   return var

> main = do
>   vty <- mkVty
>   ichan <- newTChanIO
>   forkIO $ pollForEvents vty ichan
>   nerd <- newTVarIO False
>   fdgds <- readConfig
>   vs <- mapM (spawnFeed nerd . fst) fdgds
>   let ui = UI { vt = vty, evchan = ichan,
>                 draw = menuDraw, handleEvt = menuEvt,
>                 fidx = 0, nr = nerd, fds = vs, guids = map snd fdgds, iidx = 0,
>                 prompt = Nothing, nf = "" }
>   runUI ui

vim:set sw=2 ts=2 et:
